'''
steady state algorithm
'''
from random import sample


def ktournament(population, k):
    '''
    randomly select k-individuals from the population
    and sort them by penalty in the increasing order
    '''
    enumerated_population = [(individual, idx) for idx, individual in
                             enumerate(population)]
    individuals, idxs = zip(*sorted(sample(enumerated_population, k)))
    return individuals, idxs


def generate_population(generate, evaluate, num_individuals):
    '''
    generates num_individuals in population
    '''
    genotypes = (generate() for _ in range(num_individuals))
    population = tuple((evaluate(genotype), genotype)
                       for genotype in genotypes)

    return population


def iteration(crossover, mutate, evaluate, population):
    '''
    steady state iteration, uses 3-tournament selection
    (ensures 2-elitism)
    '''
# Selection = 3-Tournament
    individuals, idxs = ktournament(population, 3)
    parents = individuals[0:2]
    bad_idx = idxs[2]

# Crossover
    genotype1, genotype2 = (genotype for _, genotype in parents)
    children = crossover(genotype1, genotype2)
    best_genotype = min(children, key=evaluate)

# Mutation
    new_genotype = mutate(best_genotype)
    penalty = evaluate(new_genotype)

    child = (penalty, new_genotype)

# Replace bad individual
    new_population = list(population)
    new_population[bad_idx] = child

    return tuple(new_population)



'''
Synchronous Island creation for testing on a single computer
'''
import ga
from ga.selection import ktournament
from ga.utility import replace, check_best

import random
from operator import itemgetter
from itertools import izip, imap


def generate_topology(num_islands, degree):
    '''
    Creates a directed, regular, fully connected graph with
    given degree of each vertex/island. This function is used
    for creating migration topology, return its adjacency list.
    '''
    island_ids = range(num_islands)
    neighbors = [island_ids[-i:] + island_ids[:-i]
                 for i in range(degree, 0, -1)]

    return list(izip(*neighbors))


def create_airports(islands, degree, migration_size, migration_interval):
    ''' adjacency matrix is given for airport topology '''

    airports = [{'migration_size': migration_size,
                 'migration_interval': migration_interval,
                 'individuals': list()}
                for _ in islands]

    topology = generate_topology(len(islands), degree)

    for island, airport, destination_idxs in izip(islands, airports, topology):
        airport['destinations'] = [airports[i] for i in destination_idxs]
        island['airport'] = airport


def receive_immigration(island):
    ''' immigration policy of some population '''
    population = island['population']
    immigrants = island['airport']['individuals']
    while immigrants:
        immigrant = immigrants.pop()
        _, bad = ktournament(population, 2)
        replace(bad, immigrant)
        check_best(island, immigrant)


def send_emmigration(island):
    '''
    randomly choses individuals for emmigration, its up to
    the programmer to schedule the migration interval
    '''
    population = island['population']
    destinations = island['airport']['destinations']
    emmigrants = random.sample(population, island['airport']['migration_size'])
    for emmigrant in emmigrants:
        destination = random.choice(destinations)
        destination['individuals'].append(emmigrant)


def create(operators, num_islands, population_size=20,
           degree=1, migration_size=1, migration_interval=200):
    ''' create archipelago, defaults to ring topology (degree=1) '''

    islands = [ga.island.create(operators, population_size)
               for _ in range(num_islands)]

    create_airports(islands, degree, migration_size, migration_interval)

    return islands


def get_solution(islands, numiters=1000):
    '''
    run synchronous genetic algorithm on all islands
    and get a solution
    '''
    for iteration in range(numiters):
        for island in islands:
            print "iteration =", iteration,
            print "penalty =", island['best']['penalty']
            receive_immigration(island)
            ga.steady.iteration(island)
            if island['stats']['born'] % island['airport']['migration_interval'] == 0:
                send_emmigration(island)

    return min((island['best'] for island in islands),
               key=itemgetter('penalty'))


def get_stagnation_solution(islands, max_stagnation=1000):
    '''
    run synchronous genetic algorithm on all islands
    and get a solution
    '''
    stagnation = 0
    penalty = float('inf')
    while stagnation < max_stagnation:
        stagnation += 1
        for island in islands:
            receive_immigration(island)
            ga.steady.iteration(island)
            if island['stats']['born'] % island['airport']['migration_interval'] == 0:
                send_emmigration(island)

            next_penalty = min(island['best']['penalty'] for island in islands)
            if next_penalty < penalty:
                penalty = next_penalty
                stagnation = 0

    return min((island['best'] for island in islands),
               key=itemgetter('penalty'))



from pyislands.permutation.mutation import reversed_sequence_mutation
from pyislands.permutation.mutation import every_city_mutation
from pyislands.permutation.mutation import every_city_mutation2

from functools import partial


def check_mutation(mutation):
    genotype = range(10)

    new_genotype = mutation(genotype)

    for i in genotype:
        assert i in new_genotype


def test_rsm():
    check_mutation(partial(reversed_sequence_mutation, 1.0))


def test_ecm():
    check_mutation(partial(every_city_mutation, 1.0))


def test_ecm2():
    check_mutation(partial(every_city_mutation2, 1.0))



from pyislands.permutation.crossover import cyclic_crossover
from pyislands.permutation.crossover import partially_mapped_crossover


def test_cx1():
    parent1 = (8, 4, 7, 3, 6, 2, 5, 1, 9, 0)
    parent2 = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9)

    child1, child2 = cyclic_crossover(parent1, parent2)

    assert child1 == (8, 1, 2, 3, 4, 5, 6, 7, 9, 0)
    assert child2 == (0, 4, 7, 3, 6, 2, 5, 1, 8, 9)


def test_cx2():
    parent1 = (0, 1, 2, 3, 4, 5, 6)
    parent2 = (6, 4, 0, 2, 1, 5, 3)

    child1, child2 = cyclic_crossover(parent1, parent2)

    assert child1 == (0, 4, 2, 3, 1, 5, 6)
    assert child2 == (6, 1, 0, 2, 4, 5, 3)


def test_pmx():
    parent1_genotype = (0, 1, 2, 3, 4, 5, 6)
    parent2_genotype = (6, 4, 0, 2, 1, 5, 3)

    children_genotypes = partially_mapped_crossover(parent1_genotype, parent2_genotype)

    for genotype in children_genotypes:
        for i in parent1_genotype:
            assert i in genotype
